@@ -14,26 +14,34 @@ It sports a small canvas and some trivial operations:
 
 """
 
-__version__ = "$Revision: 2214 $"
-# $HeadURL: http://svn.devjavu.com/gaphor/gaphas/tags/gaphas-0.3.6/demo.py $
+__version__ = "$Revision$"
+# $HeadURL$
 
-import pygtk
-pygtk.require('2.0') 
+try:
+    import pygtk
+except ImportError:
+    pass
+else:
+    pygtk.require('2.0') 
 
 import math
 import gtk
 import cairo
 from gaphas import Canvas, GtkView, View
-from gaphas.examples import Box, Text, FatLine, Circle, DefaultExampleTool
+from gaphas.examples import Box, PortoBox, Text, FatLine, Circle
 from gaphas.item import Line, NW, SE
 from gaphas.tool import PlacementTool, HandleTool
+from gaphas.aspect import Segment
 from gaphas.painter import ItemPainter
 from gaphas import state
-from gaphas.util import text_extents
+from gaphas.util import text_extents, text_underline
 
 from gaphas import painter
 #painter.DEBUG_DRAW_BOUNDING_BOX = True
 
+# Ensure data gets picked well:
+import gaphas.picklers
+
 # Global undo list
 undo_list = []
 
@@ -95,9 +103,15 @@ class MyText(Text):
         cr.stroke()
 
 
+class UnderlineText(Text):
+
+    def draw(self, context):
+        cr = context.cairo
+        text_underline(cr, 0, 0, "Some text(y)")
+ 
+
 def create_window(canvas, title, zoom=1.0):
     view = GtkView()
-    view.tool = DefaultExampleTool()
 
     w = gtk.Window()
     w.set_title(title)
@@ -119,7 +133,7 @@ def create_window(canvas, title, zoom=1.0):
 
     def on_clicked(button, view):
         #view.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.CROSSHAIR))
-        view.tool.grab(PlacementTool(factory(view, MyBox), HandleTool(), 2))
+        view.tool.grab(PlacementTool(view, factory(view, MyBox), HandleTool(), 2))
 
     b.connect('clicked', on_clicked, view)
     v.add(b)
@@ -127,7 +141,7 @@ def create_window(canvas, title, zoom=1.0):
     b = gtk.Button('Add line')
 
     def on_clicked(button):
-        view.tool.grab(PlacementTool(factory(view, MyLine), HandleTool(), 1))
+        view.tool.grab(PlacementTool(view, factory(view, MyLine), HandleTool(), 1))
 
     b.connect('clicked', on_clicked)
     v.add(b)
@@ -156,8 +170,9 @@ def create_window(canvas, title, zoom=1.0):
 
     def on_clicked(button):
         if isinstance(view.focused_item, Line):
-            view.focused_item.split_segment(0)
-            view.queue_draw_item(view.focused_item, handles=True)
+            segment = Segment(view.focused_item, view)
+            segment.split_segment(0)
+            view.queue_draw_item(view.focused_item)
 
     b.connect('clicked', on_clicked)
     v.add(b)
@@ -167,7 +182,7 @@ def create_window(canvas, title, zoom=1.0):
     def on_clicked(button):
         if view.focused_item:
             canvas.remove(view.focused_item)
-            print 'items:', canvas.get_all_items()
+            #print 'items:', canvas.get_all_items()
 
     b.connect('clicked', on_clicked)
     v.add(b)
@@ -194,14 +209,16 @@ def create_window(canvas, title, zoom=1.0):
         global undo_list
         apply_me = list(undo_list)
         del undo_list[:]
+        print 'Actions on the undo stack:', len(apply_me)
         apply_me.reverse()
         saveapply = state.saveapply
         for event in apply_me:
             print 'Undo: invoking', event
             saveapply(*event)
+            print 'New undo stack size:', len(undo_list)
             # Visualize each event:
-            while gtk.events_pending():
-                gtk.main_iteration()
+            #while gtk.events_pending():
+            #    gtk.main_iteration()
 
     b.connect('clicked', on_clicked)
     v.add(b)
@@ -271,6 +288,64 @@ def create_window(canvas, title, zoom=1.0):
     b.connect('clicked', on_clicked, [0])
     v.add(b)
 
+
+    b = gtk.Button('Pickle (save)')
+
+    def on_clicked(button, li):
+        f = open('demo.pickled', 'w')
+        try:
+            import cPickle as pickle
+            pickle.dump(view.canvas, f)
+        finally:
+            f.close()
+
+    b.connect('clicked', on_clicked, [0])
+    v.add(b)
+
+
+    b = gtk.Button('Unpickle (load)')
+
+    def on_clicked(button, li):
+        f = open('demo.pickled', 'r')
+        try:
+            import cPickle as pickle
+            canvas = pickle.load(f)
+            canvas.update_now()
+        finally:
+            f.close()
+        create_window(canvas, 'Unpickled diagram')
+
+    b.connect('clicked', on_clicked, [0])
+    v.add(b)
+
+
+    b = gtk.Button('Unpickle (in place)')
+
+    def on_clicked(button, li):
+        f = open('demo.pickled', 'r')
+        try:
+            import cPickle as pickle
+            canvas = pickle.load(f)
+        finally:
+            f.close()
+        #[i.request_update() for i in canvas.get_all_items()]
+        canvas.update_now()
+        view.canvas = canvas
+
+    b.connect('clicked', on_clicked, [0])
+    v.add(b)
+
+
+    b = gtk.Button('Reattach (in place)')
+
+    def on_clicked(button, li):
+        view.canvas = None
+        view.canvas = canvas
+
+    b.connect('clicked', on_clicked, [0])
+    v.add(b)
+
+
     # Add the actual View:
 
     t = gtk.Table(2,2)
@@ -296,13 +371,10 @@ def create_window(canvas, title, zoom=1.0):
     view.connect('hover-changed', handle_changed, 'hover')
     view.connect('selection-changed', handle_changed, 'selection')
 
-def main():
-    c=Canvas()
-
-    create_window(c, 'View created before')
-
-    # Add stuff to the canvas:
-
+    
+def create_canvas(c=None):
+    if not c:
+        c = Canvas()
     b=MyBox()
     b.min_width = 20
     b.min_height = 30
@@ -321,7 +393,6 @@ def main():
     fl.matrix.translate(100, 100)
     c.add(fl)
 
-
     circle = Circle()
     h1, h2 = circle.handles()
     circle.radius = 20
@@ -329,7 +400,7 @@ def main():
     c.add(circle)
 
     # AJM: extra boxes:
-    bb=Box()
+    bb = Box()
     print 'box', bb
     bb.matrix.rotate(math.pi/4.)
     c.add(bb, parent=b)
@@ -339,15 +410,26 @@ def main():
 #        bb.matrix.rotate(math.pi/4.0 * i / 10.0)
 #        c.add(bb, parent=b)
 
-    t=MyText('Single line')
+    b = PortoBox(60, 60)
+    b.min_width = 40
+    b.min_height = 50
+    b.matrix.translate(55, 55)
+    c.add(b)
+
+    t = UnderlineText()
+    t.matrix.translate(70,30)
+    c.add(t)
+
+    t = MyText('Single line')
     t.matrix.translate(70,70)
     c.add(t)
 
-    l=MyLine()
-    l.handles()[1].pos = (30, 30)
-    l.split_segment(0, 3)
-    l.matrix.translate(30, 60)
+    l = MyLine()
     c.add(l)
+    l.handles()[1].pos = (30, 30)
+    segment = Segment(l, view=None)
+    segment.split_segment(0, 3)
+    l.matrix.translate(30, 60)
     l.orthogonal = True
 
     off_y = 0
@@ -363,6 +445,10 @@ def main():
     t.matrix.translate(70,100)
     c.add(t)
 
+    return c
+
+
+def main():
     ##
     ## State handling (a.k.a. undo handlers)
     ##
@@ -373,6 +459,12 @@ def main():
     def print_handler(event):
         print 'event:', event
 
+    c=Canvas()
+
+    create_window(c, 'View created before')
+
+    create_canvas(c)
+
     #state.subscribers.add(print_handler)
 
     ##
@@ -383,6 +475,7 @@ def main():
 
     gtk.main()
 
+
 if __name__ == '__main__':
     import sys
     if '-p' in sys.argv:
