@@ -1,18 +1,65 @@
 """
 A Canvas owns a set of Items and acts as a container for both the items
 and a constraint solver.
+
+Connections
+===========
+
+Getting Connection Information
+==============================
+To get connected item to a handle::
+
+    c = canvas.get_connection(handle)
+    if c != None:
+        print c.connected
+        print c.port
+        print c.constraint
+
+
+To get all connected items (i.e. items on both sides of a line)::
+
+    classes = (i.connected for i in canvas.get_connections(item=line))
+
+
+To get connecting items (i.e. all lines connected to a class)::
+
+    lines = (c.item for c in canvas.get_connections(connected=item))
+
 """
 
-__version__ = "$Revision: 2239 $"
-# $HeadURL: http://svn.devjavu.com/gaphor/gaphas/tags/gaphas-0.3.6/gaphas/canvas.py $
+__version__ = "$Revision$"
+# $HeadURL$
+
+from collections import namedtuple
+import logging
 
-import cairo
 from cairo import Matrix
 from gaphas import tree
 from gaphas import solver
+from gaphas import table
 from gaphas.decorators import nonrecursive, async, PRIORITY_HIGH_IDLE
 from state import observed, reversible_method, reversible_pair
-from gaphas.sorter import Sorter
+
+
+#
+# Information about two connected items
+#
+# - item: connecting item
+# - handle: handle of connecting item (points connected item)
+# - connected: connected item
+# - port: port of connected item
+# - constraint: optional connection constraint
+# - callback: optional disconnection callback
+#
+Connection = namedtuple('Connection',
+        'item handle connected port constraint callback')
+
+
+class ConnectionError(Exception):
+    """
+    Exception raised when there is an error when connecting an items with
+    each other.
+    """
 
 
 class Context(object):
@@ -28,6 +75,7 @@ class Context(object):
     ... except: 'got exc'
     'got exc'
     """
+
     def __init__(self, **kwargs):
         self.__dict__.update(**kwargs)
 
@@ -43,20 +91,18 @@ class Canvas(object):
     def __init__(self):
         self._tree = tree.Tree()
         self._solver = solver.Solver()
+        self._connections = table.Table(Connection, range(4))
         self._dirty_items = set()
         self._dirty_matrix_items = set()
+        self._dirty_index = False
 
         self._registered_views = set()
-
-        self._sorter = Sorter(self._tree)
-
-    sorter = property(lambda s: s._sorter)
     
     solver = property(lambda s: s._solver)
 
 
     @observed
-    def add(self, item, parent=None):
+    def add(self, item, parent=None, index=None):
         """
         Add an item to the canvas.
 
@@ -70,12 +116,13 @@ class Canvas(object):
         True
         """
         assert item not in self._tree.nodes, 'Adding already added node %s' % item
-        self._tree.add(item, parent)
-        item._set_canvas(self)
-        item._sort_key = self.sorter.get_key(item)
+        self._tree.add(item, parent, index)
+        self._dirty_index = True
 
         self.update_matrix(item, parent)
 
+        item._set_canvas(self)
+
         self.request_update(item)
 
 
@@ -85,7 +132,6 @@ class Canvas(object):
         Remove is done in a separate, @observed, method so the undo system
         can restore removed items in the right order.
         """
-        self.remove_connections_to_item(item)
         item._set_canvas(None)
         self._tree.remove(item)
         self._update_views(removed_items=(item,))
@@ -108,31 +154,26 @@ class Canvas(object):
         """
         for child in reversed(self.get_children(item)):
             self.remove(child)
+        self.remove_connections_to_item(item)
         self._remove(item)
 
     reversible_pair(add, _remove,
-                    bind1={'parent': lambda self, item: self.get_parent(item)})
+                    bind1={'parent': lambda self, item: self.get_parent(item),
+                           'index': lambda self, item: self._tree.get_siblings(item).index(item) })
 
 
-    def remove_connections_to_item(self, item):
-        """
-        Remove all connections (handles connected to and constraints)
-        for a specific item.
-        This is some brute force cleanup (e.g. if constraints are referenced
-        by items, those references are not cleaned up).
-        """
-        for i, h in self.get_connected_items(item):
-            h.disconnect()
-            # Never mind..
-            h.connected_to = None
-            h.disconnect = lambda: 0
-
-
-    def reparent(self, item, parent):
+    @observed
+    def reparent(self, item, parent, index=None):
         """
         Set new parent for an item.
         """
-        self._tree.reparent(item, parent)
+        self._tree.reparent(item, parent, index)
+
+        self._dirty_index = True
+
+    reversible_method(reparent, reverse=reparent,
+                      bind={'parent': lambda self, item: self.get_parent(item),
+                            'index': lambda self, item: self._tree.get_siblings(item).index(item) })
 
 
     def get_all_items(self):
@@ -252,10 +293,176 @@ class Canvas(object):
         return self._tree.get_all_children(item)
 
 
-    def get_connected_items(self, item):
+    @observed
+    def connect_item(self, item, handle, connected, port, constraint=None, callback=None):
         """
-        Return a set of items that are connected to ``item``.
-        The list contains tuples (item, handle). As a result an item may be
+        Create a connection between two items. The connection is registered
+        and the constraint is added to the constraint solver.
+
+        The pair (item, handle) should be unique and not yet connected.
+
+        The callback is invoked when the connection is broken.
+
+        :Parameters:
+         item
+            Connecting item (i.e. a line).
+         handle
+            Handle of connecting item.
+         connected
+            Connected item (i.e. a box).
+         port
+            Port of connected item.
+         constraint
+            Constraint to keep the connection in place.
+         callback
+            Function to be called on disconnection.
+
+        ConnectionError is raised in case handle is already registered on a
+        connection.
+        """
+        if self.get_connection(handle):
+            raise ConnectionError('Handle %r of item %r is already connected' % (handle, item))
+
+        self._connections.insert(item, handle, connected, port, constraint, callback)
+
+        if constraint:
+            self._solver.add_constraint(constraint)
+
+
+    def disconnect_item(self, item, handle=None):
+        """
+        Disconnect the connections of an item. If handle is not None, only the
+        connection for that handle is disconnected.
+        """
+        # disconnect on canvas level
+        for cinfo in list(self._connections.query(item=item, handle=handle)):
+            self._disconnect_item(*cinfo)
+
+
+    @observed
+    def _disconnect_item(self, item, handle, connected, port, constraint, callback):
+        """
+        Perform the real disconnect.
+        """
+        # Same arguments as connect_item, makes reverser easy
+        if constraint:
+            self._solver.remove_constraint(constraint)
+
+        if callback:
+            callback()
+
+        self._connections.delete(item, handle, connected, port, constraint, callback)
+
+    reversible_pair(connect_item, _disconnect_item)
+
+
+    def remove_connections_to_item(self, item):
+        """
+        Remove all connections (handles connected to and constraints)
+        for a specific item (to and from the item).
+        This is some brute force cleanup (e.g. if constraints are referenced
+        by items, those references are not cleaned up).
+        """
+        disconnect_item = self._disconnect_item
+        # remove connections from this item
+        for cinfo in list(self._connections.query(item=item)):
+            disconnect_item(*cinfo)
+        # remove constraints to this item
+        for cinfo in list(self._connections.query(connected=item)):
+            disconnect_item(*cinfo)
+    
+
+    @observed
+    def reconnect_item(self, item, handle, constraint=None):
+        """
+        Update an existing connection. This is mainly useful to provide a new
+        constraint or callback to the connection. ``item`` and ``handle`` are
+        the keys to the to-be-updated connection.
+
+        >>> c = Canvas()
+        >>> from gaphas import item
+        >>> i = item.Line()
+        >>> c.add(i)
+        >>> ii = item.Line()
+        >>> c.add(ii, i)
+        >>> iii = item.Line()
+        >>> c.add(iii, ii)
+
+        We need a few constraints, because that's what we're updating:
+
+        >>> from gaphas.constraint import EqualsConstraint
+        >>> cons1 = EqualsConstraint(i.handles()[0].pos.x, i.handles()[0].pos.x)
+        >>> cons2 = EqualsConstraint(i.handles()[0].pos.y, i.handles()[0].pos.y)
+        >>> c.connect_item(i, i.handles()[0], ii, ii.ports()[0], cons1)
+        >>> c.get_connection(i.handles()[0]) # doctest: +ELLIPSIS
+        Connection(item=<gaphas.item.Line object at 0x...)
+        >>> c.get_connection(i.handles()[0]).constraint is cons1
+        True
+        >>> cons1 in c.solver.constraints
+        True
+        >>> c.reconnect_item(i, i.handles()[0], cons2, lambda: 0)
+        >>> c.get_connection(i.handles()[0]) # doctest: +ELLIPSIS
+        Connection(item=<gaphas.item.Line object at 0x...)
+        >>> c.get_connection(i.handles()[0]).constraint is cons2
+        True
+        >>> cons1 in c.solver.constraints
+        False
+        >>> cons2 in c.solver.constraints
+        True
+
+        An exception is raised if no connection exists:
+        >>> c.reconnect_item(ii, ii.handles()[0], cons2, lambda: 0) # doctest: +ELLIPSIS
+        Traceback (most recent call last):
+        ...
+        ValueError: No data available for item ...
+
+        """
+        # checks:
+        cinfo = self.get_connection(handle)
+        if not cinfo:
+            raise ValueError, 'No data available for item "%s" and handle "%s"' % (item, handle)
+
+        if cinfo.constraint:
+            self._solver.remove_constraint(cinfo.constraint)
+        self._connections.delete(item=cinfo.item, handle=cinfo.handle)
+
+        self._connections.insert(item, handle, cinfo.connected, cinfo.port, constraint, cinfo.callback)
+        if constraint:
+            self._solver.add_constraint(constraint)
+
+    reversible_method(reconnect_item, reverse=reconnect_item,
+                      bind={'constraint': lambda self, item, handle: self.get_connection(handle).constraint })
+
+
+    def get_connection(self, handle):
+        """
+        Get connection information for specified handle.
+
+        >>> c = Canvas()
+        >>> from gaphas.item import Line
+        >>> line = Line()
+        >>> from gaphas import item
+        >>> i = item.Line()
+        >>> c.add(i)
+        >>> ii = item.Line()
+        >>> c.add(ii)
+        >>> c.connect_item(i, i.handles()[0], ii, ii.ports()[0])
+        >>> c.get_connection(i.handles()[0])     # doctest: +ELLIPSIS
+        Connection(item=<gaphas.item.Line object at 0x...)
+        >>> c.get_connection(i.handles()[1])     # doctest: +ELLIPSIS
+        >>> c.get_connection(ii.handles()[0])    # doctest: +ELLIPSIS
+        """
+        try:
+            return self._connections.query(handle=handle).next()
+        except StopIteration, ex:
+            return None
+
+
+    def get_connections(self, item=None, handle=None, connected=None, port=None):
+        """
+        Return an iterator of connection information.
+
+        The list contains  (item, handle). As a result an item may be
         in the list more than once (depending on the number of handles that
         are connected). If ``item`` is connected to itself it will also appear
         in the list.
@@ -268,24 +475,49 @@ class Canvas(object):
         >>> c.add(ii)
         >>> iii = item.Line()
         >>> c.add (iii)
-        >>> i.handles()[0].connected_to = ii
-        >>> list(c.get_connected_items(i))
+        >>> c.connect_item(i, i.handles()[0], ii, ii.ports()[0], None)
+
+        >>> list(c.get_connections(item=i)) # doctest: +ELLIPSIS
+        [Connection(item=<gaphas.item.Line object at 0x...]
+        >>> list(c.get_connections(connected=i))
         []
-        >>> ii.handles()[0].connected_to = iii
-        >>> list(c.get_connected_items(ii)) # doctest: +ELLIPSIS
-        [(<gaphas.item.Line ...>, <Handle object on (0, 0)>)]
-        >>> list(c.get_connected_items(iii)) # doctest: +ELLIPSIS
-        [(<gaphas.item.Line ...>, <Handle object on (0, 0)>)]
+        >>> list(c.get_connections(connected=ii)) # doctest: +ELLIPSIS
+        [Connection(item=<gaphas.item.Line object at 0x...]
+
+        >>> c.connect_item(ii, ii.handles()[0], iii, iii.ports()[0], None)
+        >>> list(c.get_connections(item=ii)) # doctest: +ELLIPSIS
+        [Connection(item=<gaphas.item.Line object at 0x...]
+        >>> list(c.get_connections(connected=iii)) # doctest: +ELLIPSIS
+        [Connection(item=<gaphas.item.Line object at 0x...]
         """
-        connected_items = set()
-        for i in self.get_all_items():
-            for h in i.handles():
-                if h.connected_to is item:
-                    connected_items.add((i, h))
-        return connected_items
+        return self._connections.query(item=item,
+                handle=handle,
+                connected=connected,
+                port=port)
 
+        
+    def sort(self, items, reverse=False):
+        """
+        Sort a list of items in the order in which they are traversed in
+        the canvas (Depth first).
+
+        >>> c = Canvas()
+        >>> from gaphas import item
+        >>> i1 = item.Line()
+        >>> c.add(i1)
+        >>> i2 = item.Line()
+        >>> c.add(i2)
+        >>> i3 = item.Line()
+        >>> c.add (i3)
+        >>> c.update() # ensure items are indexed
+        >>> i1._canvas_index
+        0
+        >>> s = c.sort([i2, i3, i1])
+        >>> s[0] is i1 and s[1] is i2 and s[2] is i3
+        True
+        """
+        return self._tree.sort(items, index_key='_canvas_index', reverse=reverse)
 
-    #{ Matrices
 
     def get_matrix_i2c(self, item, calculate=False):
         """
@@ -313,8 +545,15 @@ class Canvas(object):
             self.update_matrix(item)
         return item._matrix_c2i
 
-
-    #{ Update cycle
+    def get_matrix_i2i(self, from_item, to_item, calculate=False):
+        i2c = self.get_matrix_i2c(from_item, calculate)
+        c2i = self.get_matrix_c2i(to_item, calculate)
+        try:
+            return i2c.multiply(c2i)
+        except AttributeError:
+            # Fall back to old behaviour
+            return i2c * c2i
+        
 
     @observed
     def request_update(self, item, update=True, matrix=True):
@@ -380,34 +619,37 @@ class Canvas(object):
 
     def _pre_update_items(self, items, cr):
         context_map = dict()
+        c = Context(cairo=cr)
         for item in items:
-            c = Context(cairo=cr)
-            try:
-                item.pre_update(c)
-            except Exception, e:
-                print 'Error while pre-updating item %s' % item
-                import traceback
-                traceback.print_exc()
+            item.pre_update(c)
 
 
     def _post_update_items(self, items, cr):
+        c = Context(cairo=cr)
         for item in items:
-            c = Context(cairo=cr)
-            try:
-                item.post_update(c)
-            except Exception, e:
-                print 'Error while updating item %s' % item
-                import traceback
-                traceback.print_exc()
+            item.post_update(c)
 
 
+    def _extend_dirty_items(self, dirty_items):
+        # item's can be marked dirty due to external constraints solving
+        if self._dirty_items:
+            dirty_items.extend(self._dirty_items)
+            self._dirty_items.clear()
+
+            dirty_items = self.sort(set(dirty_items), reverse=True)
+
     @nonrecursive
     def update_now(self):
         """
         Peform an update of the items that requested an update.
         """
 
-        sort = self._sorter.sort
+        if self._dirty_index:
+            self.update_index()
+            self._dirty_index = False
+
+        sort = self.sort
+        extend_dirty_items = self._extend_dirty_items
 
         # perform update requests for parents of dirty items
         dirty_items = self._dirty_items
@@ -423,35 +665,20 @@ class Canvas(object):
             cr = self._obtain_cairo_context()
 
             # allow programmers to perform tricks and hacks before item
-            # full update
+            # full update (only called for items that requested a full update)
             self._pre_update_items(dirty_items, cr)
 
             # recalculate matrices
             dirty_matrix_items = self.update_matrices(self._dirty_matrix_items)
             self._dirty_matrix_items.clear()
 
-            # request solving of external constraints associated with dirty
-            # items
-            request_resolve = self._solver.request_resolve
-            for item in dirty_matrix_items:
-                for h in item.handles():
-                    request_resolve(h.x, projections_only=True)
-                    request_resolve(h.y, projections_only=True)
-            
-
-            # solve all constraints
-            self._solver.solve()
+            self.update_constraints(dirty_matrix_items)
 
             # no matrix can change during constraint solving
             assert not self._dirty_matrix_items, 'No matrices may have been marked dirty (%s)' % (self._dirty_matrix_items,)
 
-            # some item's can be marked dirty due to external constraints
-            # solving
-            if self._dirty_items:
-                dirty_items.extend(self._dirty_items)
-                self._dirty_items.clear()
-
-                dirty_items = sort(set(dirty_items), reverse=True)
+            # item's can be marked dirty due to external constraints solving
+            extend_dirty_items(dirty_items)
 
             assert not self._dirty_items, 'No items may have been marked dirty (%s)' % (self._dirty_items,)
 
@@ -462,12 +689,18 @@ class Canvas(object):
             # recalculate matrices of normalized items
             dirty_matrix_items.update(self.update_matrices(normalized_items))
 
+            # ensure constraints are still true after normalization
+            self._solver.solve()
+
+            # item's can be marked dirty due to normalization and solving
+            extend_dirty_items(dirty_items)
+
+            assert not self._dirty_items, 'No items may have been marked dirty (%s)' % (self._dirty_items,)
+
             self._post_update_items(dirty_items, cr)
 
         except Exception, e:
-            print 'Error while updating canvas'
-            import traceback
-            traceback.print_exc()
+            logging.error('Error while updating canvas', e)
 
         assert len(self._dirty_items) == 0 and len(self._dirty_matrix_items) == 0, \
                 'dirty: %s; matrix: %s' % (self._dirty_items, self._dirty_matrix_items)
@@ -511,7 +744,11 @@ class Canvas(object):
         item._matrix_i2c = Matrix(*item.matrix)
 
         if parent is not None:
-            item._matrix_i2c *= parent._matrix_i2c
+            try:
+                item._matrix_i2c = item._matrix_i2c.multiply(parent._matrix_i2c)
+            except AttributeError:
+                # Fall back to old behaviour
+                item._matrix_i2c *= parent._matrix_i2c
 
         if orig_matrix_i2c is None or orig_matrix_i2c != item._matrix_i2c:
             # calculate c2i matrix and view matrices
@@ -519,6 +756,22 @@ class Canvas(object):
             item._matrix_c2i.invert()
 
 
+    def update_constraints(self, items):
+        """
+        Update constraints. Also variables may be marked as dirty before the
+        constraint solver kicks in.
+        """
+        # request solving of external constraints associated with dirty items
+        request_resolve = self._solver.request_resolve
+        for item in items:
+            for p in item._canvas_projections:
+                request_resolve(p[0], projections_only=True)
+                request_resolve(p[1], projections_only=True)
+
+        # solve all constraints
+        self._solver.solve()
+
+
     def _normalize(self, items):
         """
         Update handle positions of items, so the first handle is always
@@ -529,7 +782,7 @@ class Canvas(object):
 
         For example having an item
 
-        >>> from item import Element
+        >>> from gaphas.item import Element
         >>> c = Canvas()
         >>> e = Element()
         >>> c.add(e)
@@ -540,7 +793,7 @@ class Canvas(object):
 
         and moving its first handle a bit
 
-        >>> e.handles()[0].x += 1
+        >>> e.handles()[0].pos.x += 1
         >>> map(float, e.handles()[0].pos)
         [1.0, 0.0]
 
@@ -553,25 +806,19 @@ class Canvas(object):
         """
         dirty_matrix_items = set()
         for item in items:
-            handles = item.handles()
-            if not handles:
-                continue
-            x, y = map(float, handles[0].pos)
-            if x:
-                item.matrix._matrix.translate(x, 0)
+            if item.normalize():
                 dirty_matrix_items.add(item)
-                for h in handles:
-                    h.x._value -= x
-            if y:
-                item.matrix._matrix.translate(0, y)
-                dirty_matrix_items.add(item)
-                for h in handles:
-                    h.y._value -= y
-
+                
         return dirty_matrix_items
 
 
-    #{ Views
+    def update_index(self):
+        """
+        Provide each item in the canvas with an index attribute. This makes
+        for fast searching of items.
+        """
+        self._tree.index_nodes('_canvas_index')
+
 
     def register_view(self, view):
         """
@@ -616,10 +863,58 @@ class Canvas(object):
             except AttributeError:
                 pass
         else:
+            import cairo
             surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
             return cairo.Context(surface)
 
 
+    def __getstate__(self):
+        """
+        Persist canvas. Dirty item sets and views are not saved.
+        """
+        d = dict(self.__dict__)
+        for n in ('_dirty_items', '_dirty_matrix_items', '_dirty_index', '_registered_views'):
+            try:
+                del d[n]
+            except KeyError:
+                pass
+        return d
+
+
+    def __setstate__(self, state):
+        """
+        Load persisted state.
+
+        Before loading the state, the constructor is called.
+        """
+        self.__dict__.update(state)
+        self._dirty_items = set(self._tree.nodes)
+        self._dirty_matrix_items = set(self._tree.nodes)
+        self._dirty_index = True
+        self._registered_views = set()
+        #self.update()
+
+
+    def project(self, item, *points):
+        """
+        Project item's points into canvas coordinate system.
+
+        If there is only one point returned than projected point is
+        returned. If there are more than one points, then tuple of
+        projected points is returned.
+        """
+        def reg(cp):
+            item._canvas_projections.add(cp)
+            return cp
+
+        if len(points) == 1:
+            return reg(CanvasProjection(points[0], item))
+        elif len(points) > 1:
+            return tuple(reg(CanvasProjection(p, item)) for p in points)
+        else:
+            raise AttributeError('There should be at least one point specified')
+
+
 class VariableProjection(solver.Projection):
     """
     Project a single `solver.Variable` to another space/coordinate system.
@@ -660,7 +955,7 @@ class CanvasProjection(object):
     (Projections).
 
     >>> canvas = Canvas()
-    >>> from item import Element
+    >>> from gaphas.item import Element
     >>> a = Element()
     >>> canvas.add(a)
     >>> a.matrix.translate(30, 2)
@@ -670,14 +965,14 @@ class CanvasProjection(object):
     cairo.Matrix(1, 0, 0, 1, 30, 2)
     >>> p = CanvasProjection(a.handles()[2].pos, a)
     >>> a.handles()[2].pos
-    (Variable(10, 40), Variable(10, 40))
+    <Position object on (10, 10)>
     >>> p[0].value
     40.0
     >>> p[1].value
     12.0
     >>> p[0].value = 63
     >>> p._point
-    (Variable(33, 40), Variable(10, 40))
+    <Position object on (33, 10)>
 
     When the variables are retrieved, new values are calculated.
     """
@@ -689,13 +984,13 @@ class CanvasProjection(object):
     def _on_change_x(self, value):
         item = self._item
         self._px = value
-        self._point[0].value, self._point[1].value = item.canvas.get_matrix_c2i(item).transform_point(value, self._py)
+        self._point.x.value, self._point.y.value = item.canvas.get_matrix_c2i(item).transform_point(value, self._py)
         item.canvas.request_update(item, matrix=False)
 
     def _on_change_y(self, value):
         item = self._item
         self._py = value
-        self._point[0].value, self._point[1].value = item.canvas.get_matrix_c2i(item).transform_point(self._px, value)
+        self._point.x.value, self._point.y.value = item.canvas.get_matrix_c2i(item).transform_point(self._px, value)
         item.canvas.request_update(item, matrix=False)
 
     def _get_value(self):
@@ -704,20 +999,21 @@ class CanvasProjection(object):
         a value attribute with the real value.
         """
         item = self._item
-        x, y = self._point
+        x, y = self._point.x, self._point.y
         self._px, self._py = item.canvas.get_matrix_i2c(item).transform_point(x, y)
         return self._px, self._py
 
+    pos = property(lambda self: map(VariableProjection,
+                                    self._point, self._get_value(),
+                                    (self._on_change_x, self._on_change_y)))
+
     def __getitem__(self, key):
-        return map(VariableProjection,
-                   self._point, self._get_value(),
-                   (self._on_change_x, self._on_change_y))[key]
+        # Note: we can not use bound methods as callbacks, since that will
+        #       cause pickle to fail.
+        return self.pos[key]
         
     def __iter__(self):
-        return iter(map(VariableProjection,
-                        self._point, self._get_value(),
-                        (self._on_change_x, self._on_change_y)))
-
+        return iter(self.pos)
 
 
 # Additional tests in @observed methods
@@ -728,8 +1024,4 @@ __test__ = {
     }
 
 
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod(optionflags=doctest.ELLIPSIS)
-
-# vim:sw=4:et
+# vim:sw=4:et:ai
