@@ -2,126 +2,23 @@
 Basic items.
 """
 
-__version__ = "$Revision: 2266 $"
-# $HeadURL: http://svn.devjavu.com/gaphor/gaphas/tags/gaphas-0.3.6/gaphas/item.py $
+__version__ = "$Revision$"
+# $HeadURL$
 
 from math import atan2
 from weakref import WeakKeyDictionary
+try:
+    # python 3.0 (better be prepared)
+    from weakref import WeakSet
+except ImportError:
+    from weakset import WeakSet
 
 from matrix import Matrix
 from geometry import distance_line_point, distance_rectangle_point
+from connector import Handle, LinePort
 from solver import solvable, WEAK, NORMAL, STRONG, VERY_STRONG
-from constraint import EqualsConstraint, LessThanConstraint
-from state import observed, reversible_method, reversible_pair, reversible_property, disable_dispatching
-
-
-class Handle(object):
-    """
-    Handles are used to support modifications of Items.
-
-    If the handle is connected to an item, the connected_to property should
-    refer to the item. A disconnect handler should be provided that handles
-    all disconnect behavior (e.g. clean up constraints and connected_to).
-    """
-    
-    _x = solvable()
-    _y = solvable()
-
-    def __init__(self, x=0, y=0, strength=NORMAL, connectable=False, movable=True):
-        self._x = x
-        self._y = y
-        self._x.strength = strength
-        self._y.strength = strength
-
-        # Flags.. can't have enough of those
-        self._connectable = connectable
-        self._movable = movable
-        self._visible = True
-        self._connected_to = None
-
-        # User data for the connection (e.g. constraints)
-        self._connection_data = None
-        # An extra property used to disconnect the constraint. Should be set
-        # by the application.
-        self._disconnect = lambda: 0
-
-    @observed
-    def _set_x(self, x):
-        self._x = x
-
-    x = reversible_property(lambda s: s._x, _set_x, bind={'x': lambda self: float(self.x) })
-    disable_dispatching(_set_x)
-
-    @observed
-    def _set_y(self, y):
-        self._y = y
-
-    y = reversible_property(lambda s: s._y, _set_y, bind={'y': lambda self: float(self.y) })
-    disable_dispatching(_set_y)
-
-    @observed
-    def _set_connectable(self, connectable):
-        self._connectable = connectable
-
-    connectable = reversible_property(lambda s: s._connectable, _set_connectable)
-
-    @observed
-    def _set_movable(self, movable):
-        self._movable = movable
-
-    movable = reversible_property(lambda s: s._movable, _set_movable)
-
-    @observed
-    def _set_visible(self, visible):
-        self._visible = visible
-
-    visible = reversible_property(lambda s: s._visible, _set_visible)
-
-    @observed
-    def _set_connected_to(self, connected_to):
-        self._connected_to = connected_to
-
-    connected_to = reversible_property(lambda s: s._connected_to,
-                                       _set_connected_to)
-
-    @observed
-    def _set_connection_data(self, connection_data):
-        self._connection_data = connection_data
-
-    connection_data = reversible_property(lambda s: s._connection_data,
-                                          _set_connection_data)
-
-    @observed
-    def _set_disconnect(self, disconnect):
-        self._disconnect = disconnect
-
-    disconnect = reversible_property(lambda s: s._disconnect, _set_disconnect)
-
-    @observed
-    def _set_pos(self, pos):
-        """
-        Set handle position (Item coordinates).
-        """
-        self.x, self.y = pos
-
-    pos = property(lambda s: (s.x, s.y), _set_pos)
-
-    def __str__(self):
-        return '<%s object on (%g, %g)>' % (self.__class__.__name__, float(self.x), float(self.y))
-    __repr__ = __str__
-
-    def __getitem__(self, index):
-        """
-        Shorthand for returning the x(0) or y(1) component of the point.
-
-            >>> h = Handle(3, 5)
-            >>> h[0]
-            Variable(3, 20)
-            >>> h[1]
-            Variable(5, 20)
-        """
-        return (self.x, self.y)[index]
-
+from constraint import EqualsConstraint, LessThanConstraint, LineConstraint, LineAlignConstraint
+from state import observed, reversible_method, reversible_pair, reversible_property
 
 class Item(object):
     """
@@ -131,17 +28,20 @@ class Item(object):
 
     - matrix: item's transformation matrix
     - canvas: canvas, which owns an item
-    
+    - constraints: list of item constraints, automatically registered
+      when the item is added to a canvas; may be extended in subclasses
+
     Private:
 
     - _canvas:      canvas, which owns an item
     - _handles:     list of handles owned by an item
-    - _constraints: item's constraints
+    - _ports:       list of ports, connectable areas of an item
     - _matrix_i2c:  item to canvas coordinates matrix
     - _matrix_c2i:  canvas to item coordinates matrix
     - _matrix_i2v:  item to view coordinates matrices
     - _matrix_v2i:  view to item coordinates matrices
     - _sort_key:  used to sort items
+    - _canvas_projections:  used to sort items
     """
 
     def __init__(self):
@@ -149,22 +49,18 @@ class Item(object):
         self._matrix = Matrix()
         self._handles = []
         self._constraints = []
+        self._ports = []
 
         # used by gaphas.canvas.Canvas to hold conversion matrices
         self._matrix_i2c = None
-        self._matrix_2ci = None
+        self._matrix_c2i = None
 
         # used by gaphas.view.GtkView to hold item 2 view matrices (view=key)
         self._matrix_i2v = WeakKeyDictionary()
         self._matrix_v2i = WeakKeyDictionary()
+	self._canvas_projections = WeakSet()
 
-        # Used by gaphas.sorter.Sorter to order items fast.
-        self._sort_key = None
-
-
-    # _set_canvas() is not observed, since this operation is initialized by
-    # Canvas.add() and Canvas.remove()
-    #@observed
+    @observed
     def _set_canvas(self, canvas):
         """
         Set the canvas. Should only be called from Canvas.add and
@@ -177,9 +73,11 @@ class Item(object):
         if canvas:
             self.setup_canvas()
 
-    canvas = reversible_property(lambda s: s._canvas,
+    canvas = reversible_property(lambda s: s._canvas, _set_canvas,
                 doc="Canvas owning this item")
 
+    constraints = property(lambda s: s._constraints,
+                doc="Item constraints")
 
     def setup_canvas(self):
         """
@@ -196,8 +94,7 @@ class Item(object):
         Called when the canvas is unset for the item.
         This method can be used to dispose constraints.
         """
-        for h in self.handles():
-            h.disconnect()
+        self.canvas.disconnect_item(self)
 
         remove = self.canvas.solver.remove_constraint
         for c in self._constraints:
@@ -249,6 +146,40 @@ class Item(object):
         """
         pass
 
+
+    def normalize(self):
+        """
+        Update handle positions of the item, so the first handle is always
+        located at (0, 0).
+        
+        Note that, since this method basically does some housekeeping during
+        the update phase, there's no need to keep track of the changes.
+
+        Alternative implementation can also be created, e.g. set (0, 0) in
+        the center of a circle or change it depending on the location of a
+        rotation point.
+
+        Returns ``True`` if some updates have been done, ``False`` otherwise.
+
+        See ``canvas._normalize()`` for tests.
+        """
+        updated = False
+        handles = self._handles
+        if handles:
+            x, y = map(float, handles[0].pos)
+            if x:
+                self.matrix.translate(x, 0)
+                updated = True
+                for h in handles:
+                    h.pos.x -= x
+            if y:
+                self.matrix.translate(0, y)
+                updated = True
+                for h in handles:
+                    h.pos.y -= y
+        return updated
+
+
     def draw(self, context):
         """
         Render the item to a canvas view.
@@ -261,13 +192,22 @@ class Item(object):
         """
         pass
 
+    
     def handles(self):
         """
         Return a list of handles owned by the item.
         """
         return self._handles
 
-    def point(self, x, y):
+
+    def ports(self):
+        """
+        Return list of ports.
+        """
+        return self._ports
+
+
+    def point(self, pos):
         """
         Get the distance from a point (``x``, ``y``) to the item.
         ``x`` and ``y`` are in item coordinates.
@@ -275,6 +215,91 @@ class Item(object):
         pass
 
 
+    def constraint(self,
+            horizontal=None,
+            vertical=None,
+            left_of=None,
+            above=None,
+            line=None,
+            delta=0.0,
+            align=None):
+        """
+        Utility (factory) method to create item's internal constraint between
+        two positions or between a position and a line.
+
+        Position is a tuple of coordinates, i.e. ``(2, 4)``.
+
+        Line is a tuple of positions, i.e. ``((2, 3), (4, 2))``.
+
+        This method shall not be used to create constraints between
+        two different items.
+
+        Created constraint is returned.
+
+        :Parameters:
+         horizontal=(p1, p2)
+            Keep positions ``p1`` and ``p2`` aligned horizontally.
+         vertical=(p1, p2)
+            Keep positions ``p1`` and ``p2`` aligned vertically.
+         left_of=(p1, p2)
+            Keep position ``p1`` on the left side of position ``p2``.
+         above=(p1, p2)
+            Keep position ``p1`` above position ``p2``.
+         line=(p, l)
+            Keep position ``p`` on line ``l``.
+        """
+        cc = None # created constraint
+        if horizontal:
+            p1, p2 = horizontal
+            cc = EqualsConstraint(p1[1], p2[1], delta)
+        elif vertical:
+            p1, p2 = vertical
+            cc = EqualsConstraint(p1[0], p2[0], delta)
+        elif left_of:
+            p1, p2 = left_of
+            cc = LessThanConstraint(p1[0], p2[0], delta)
+        elif above:
+            p1, p2 = above
+            cc = LessThanConstraint(p1[1], p2[1], delta)
+        elif line:
+            pos, l = line
+            if align is None:
+                cc = LineConstraint(line=l, point=pos)
+            else:
+                cc = LineAlignConstraint(line=l, point=pos, align=align, delta=delta)
+        else:
+            raise ValueError('Constraint incorrectly specified')
+        assert cc is not None
+        self._constraints.append(cc)
+        return cc
+
+
+    def __getstate__(self):
+        """
+        Persist all, but calculated values (``_matrix_?2?``).
+        """
+        d = dict(self.__dict__)
+        for n in ('_matrix_i2c', '_matrix_c2i', '_matrix_i2v', '_matrix_v2i'):
+            try:
+                del d[n]
+            except KeyError:
+                pass
+        d['_canvas_projections'] = tuple(self._canvas_projections)
+        return d
+
+
+    def __setstate__(self, state):
+        """
+        Set state. No ``__init__()`` is called.
+        """
+        for n in ('_matrix_i2c', '_matrix_c2i'):
+            setattr(self, n, None)
+        for n in ('_matrix_i2v', '_matrix_v2i'):
+            setattr(self, n, WeakKeyDictionary())
+        self.__dict__.update(state)
+        self._canvas_projections = WeakSet(state['_canvas_projections'])
+
+
 [ NW,
   NE,
   SE,
@@ -293,34 +318,35 @@ class Element(Item):
         super(Element, self).__init__()
         self._handles = [ h(strength=VERY_STRONG) for h in [Handle]*4 ]
 
-        eq = EqualsConstraint
-        lt = LessThanConstraint
         handles = self._handles
         h_nw = handles[NW]
         h_ne = handles[NE]
         h_sw = handles[SW]
         h_se = handles[SE]
 
-        # create minimal size constraints
-        self._c_min_w = LessThanConstraint(smaller=h_nw.x, bigger=h_se.x, delta=10)
-        self._c_min_h = LessThanConstraint(smaller=h_nw.y, bigger=h_se.y, delta=10)
+        # edge of element define default element ports
+        self._ports = [
+            LinePort(h_nw.pos, h_ne.pos),
+            LinePort(h_ne.pos, h_se.pos), 
+            LinePort(h_se.pos, h_sw.pos), 
+            LinePort(h_sw.pos, h_nw.pos)
+        ]
 
         # setup constraints
-        self._constraints.extend([
-            eq(a=h_nw.y, b=h_ne.y),
-            eq(a=h_nw.x, b=h_sw.x),
-            eq(a=h_se.y, b=h_sw.y),
-            eq(a=h_se.x, b=h_ne.x),
-            # set h_nw < h_se constraints
-            # with minimal size functionality
-            self._c_min_w,
-            self._c_min_h,
-        ])
+        self.constraint(horizontal=(h_nw.pos, h_ne.pos))
+        self.constraint(horizontal=(h_se.pos, h_sw.pos))
+        self.constraint(vertical=(h_nw.pos, h_sw.pos))
+        self.constraint(vertical=(h_se.pos, h_ne.pos))
+
+        # create minimal size constraints
+        self._c_min_w = self.constraint(left_of=(h_nw.pos, h_se.pos), delta=10)
+        self._c_min_h = self.constraint(above=(h_nw.pos, h_se.pos), delta=10)
 
         # set width/height when minimal size constraints exist
         self.width = width
         self.height = height
 
+
     def setup_canvas(self):
         super(Element, self).setup_canvas()
 
@@ -334,15 +360,15 @@ class Element(Item):
         >>> b.width = 20
         >>> b.width
         20.0
-        >>> b._handles[NW].x
+        >>> b._handles[NW].pos.x
         Variable(0, 40)
-        >>> b._handles[SE].x
+        >>> b._handles[SE].pos.x
         Variable(20, 40)
         """
         if width < self.min_width:
             width = self.min_width
         h = self._handles
-        h[SE].x = h[NW].x + width
+        h[SE].pos.x = h[NW].pos.x + width
 
 
     def _get_width(self):
@@ -351,7 +377,7 @@ class Element(Item):
         right handle.
         """
         h = self._handles
-        return float(h[SE].x) - float(h[NW].x)
+        return float(h[SE].pos.x) - float(h[NW].pos.x)
 
     width = property(_get_width, _set_width)
 
@@ -364,22 +390,22 @@ class Element(Item):
         >>> b.height = 2
         >>> b.height
         10.0
-        >>> b._handles[NW].y
+        >>> b._handles[NW].pos.y
         Variable(0, 40)
-        >>> b._handles[SE].y
+        >>> b._handles[SE].pos.y
         Variable(10, 40)
         """
         if height < self.min_height:
             height = self.min_height
         h = self._handles
-        h[SE].y = h[NW].y + height
+        h[SE].pos.y = h[NW].pos.y + height
 
     def _get_height(self):
         """
         Height.
         """
         h = self._handles
-        return float(h[SE].y) - float(h[NW].y)
+        return float(h[SE].pos.y) - float(h[NW].pos.y)
 
     height = property(_get_height, _set_height)
 
@@ -412,13 +438,17 @@ class Element(Item):
     min_height = reversible_property(lambda s: s._c_min_h.delta, _set_min_height)
 
         
-    def point(self, x, y):
+    def point(self, pos):
         """
         Distance from the point (x, y) to the item.
+
+        >>> e = Element()
+        >>> e.point((20, 10))
+        10.0
         """
         h = self._handles
-        hnw, hse = h[NW], h[SE]
-        return distance_rectangle_point(map(float, (hnw.x, hnw.y, hse.x, hse.y)), (x, y))
+        pnw, pse = h[NW].pos, h[SE].pos
+        return distance_rectangle_point(map(float, (pnw.x, pnw.y, pse.x, pse.y)), pos)
 
 
 class Line(Item):
@@ -442,7 +472,9 @@ class Line(Item):
 
     def __init__(self):
         super(Line, self).__init__()
-        self._handles = [Handle(connectable=True), Handle(10, 10, connectable=True)]
+        self._handles = [Handle(connectable=True), Handle((10, 10), connectable=True)]
+        self._ports = []
+        self._update_ports()
 
         self._line_width = 2
         self._fuzziness = 0
@@ -474,24 +506,27 @@ class Line(Item):
 
         for c in self._orthogonal_constraints:
             self.canvas.solver.remove_constraint(c)
+        del self._orthogonal_constraints[:]
 
         if not orthogonal:
             return
 
         h = self._handles
-        if len(h) < 3:
-            self.split_segment(0)
+        #if len(h) < 3:
+        #    self.split_segment(0)
         eq = EqualsConstraint #lambda a, b: a - b
         add = self.canvas.solver.add_constraint
         cons = []
         rest = self._horizontal and 1 or 0
         for pos, (h0, h1) in enumerate(zip(h, h[1:])):
+            p0 = h0.pos
+            p1 = h1.pos
             if pos % 2 == rest: # odd
-                cons.append(add(eq(a=h0.x, b=h1.x)))
+                cons.append(add(eq(a=p0.x, b=p1.x)))
             else:
-                cons.append(add(eq(a=h0.y, b=h1.y)))
-            self.canvas.solver.request_resolve(h1.x)
-            self.canvas.solver.request_resolve(h1.y)
+                cons.append(add(eq(a=p0.y, b=p1.y)))
+            self.canvas.solver.request_resolve(p1.x)
+            self.canvas.solver.request_resolve(p1.y)
         self._set_orthogonal_constraints(cons)
         self.request_update()
 
@@ -511,6 +546,8 @@ class Line(Item):
         >>> a.orthogonal
         False
         """
+        if orthogonal and len(self.handles()) < 3:
+            raise ValueError, "Can't set orthogonal line with less than 3 handles"
         self._update_orthogonal_constraints(orthogonal)
 
     orthogonal = reversible_property(lambda s: bool(s._orthogonal_constraints), _set_orthogonal)
@@ -532,7 +569,7 @@ class Line(Item):
         False
         """
         self._inner_set_horizontal(horizontal)
-        self._update_orthogonal_constraints(self._orthogonal_constraints)
+        self._update_orthogonal_constraints(self.orthogonal)
 
     horizontal = reversible_property(lambda s: s._horizontal, _set_horizontal)
 
@@ -562,98 +599,35 @@ class Line(Item):
     reversible_pair(_reversible_insert_handle, _reversible_remove_handle, \
             bind1={'index': lambda self, handle: self._handles.index(handle)})
 
+    @observed
+    def _reversible_insert_port(self, index, port):
+        self._ports.insert(index, port)
 
-    def split_segment(self, segment, parts=2):
+    @observed
+    def _reversible_remove_port(self, port):
+        self._ports.remove(port)
+
+    reversible_pair(_reversible_insert_port, _reversible_remove_port, \
+            bind1={'index': lambda self, port: self._ports.index(port)})
+
+
+    def _create_handle(self, pos, strength=WEAK):
+        return Handle(pos, strength=strength)
+
+    
+    def _create_port(self, p1, p2):
+        return LinePort(p1, p2)
+
+
+    def _update_ports(self):
         """
-        Split one segment in the Line in ``parts`` equal pieces.
-        ``segment`` 0 is the first segment (between handles 0 and 1).
-        The min number of parts is 2.
-
-        A list of new handles is returned.
-
-        Note that ``split_segment`` is not able to reconnect constraints that
-        are connected to the segment. 
-
-        >>> a = Line()
-        >>> a.handles()[1].pos = (20, 0)
-        >>> len(a.handles())
-        2
-        >>> a.split_segment(0)
-        [<Handle object on (10, 0)>]
-        >>> a.handles()
-        [<Handle object on (0, 0)>, <Handle object on (10, 0)>, <Handle object on (20, 0)>]
-
-        A line segment can be split into multiple (equal) parts:
-
-        >>> b = Line()
-        >>> b.handles()[1].pos = (20, 16)
-        >>> b.handles()
-        [<Handle object on (0, 0)>, <Handle object on (20, 16)>]
-        >>> b.split_segment(0, parts=4)
-        [<Handle object on (5, 4)>, <Handle object on (10, 8)>, <Handle object on (15, 12)>]
-        >>> len(b.handles())
-        5
-        >>> b.handles()
-        [<Handle object on (0, 0)>, <Handle object on (5, 4)>, <Handle object on (10, 8)>, <Handle object on (15, 12)>, <Handle object on (20, 16)>]
+        Update line ports.
         """
-        assert parts >= 2
-        assert segment >= 0
-        def do_split(segment, parts):
-            h0 = self._handles[segment]
-            h1 = self._handles[segment + 1]
-            dx, dy = h1.x - h0.x, h1.y - h0.y
-            new_h = Handle(h0.x + dx / parts, h0.y + dy / parts, strength=WEAK)
-            self._reversible_insert_handle(segment + 1, new_h)
-            if parts > 2:
-                do_split(segment + 1, parts - 1)
-        do_split(segment, parts)
-        # Force orthogonal constraints to be recreated
-        self._update_orthogonal_constraints(self.orthogonal)
-        return self._handles[segment+1:segment+parts]
-
-    def merge_segment(self, segment, parts=2):
-        """
-        Merge the ``segment`` and the next.
-        The parts parameter indicates how many segments should be merged
-
-        The deleted handles are returned as a list.
-
-        >>> a = Line()
-        >>> a.handles()[1].pos = (20, 0)
-        >>> _ = a.split_segment(0)
-        >>> a.handles()
-        [<Handle object on (0, 0)>, <Handle object on (10, 0)>, <Handle object on (20, 0)>]
-        >>> a.merge_segment(0)
-        [<Handle object on (10, 0)>]
-        >>> a.handles()
-        [<Handle object on (0, 0)>, <Handle object on (20, 0)>]
-        >>> try: a.merge_segment(0)
-        ... except AssertionError: print 'okay'
-        okay
-
-        More than two segments can be merged at once:
-        >>> _ = a.split_segment(0)
-        >>> _ = a.split_segment(0)
-        >>> _ = a.split_segment(0)
-        >>> a.handles()
-        [<Handle object on (0, 0)>, <Handle object on (2.5, 0)>, <Handle object on (5, 0)>, <Handle object on (10, 0)>, <Handle object on (20, 0)>]
-        >>> a.merge_segment(0, parts=4)
-        [<Handle object on (2.5, 0)>, <Handle object on (5, 0)>, <Handle object on (10, 0)>]
-        >>> a.handles()
-        [<Handle object on (0, 0)>, <Handle object on (20, 0)>]
-        """
-        assert len(self._handles) > 2, 'Not enough segments'
-        if 0 >= segment > len(self._handles) - 1:
-            raise IndexError("index out of range (0 > %d > %d)" % (segment, len(self._handles) - 1))
-        if segment == 0: segment = 1
-        deleted_handles = self._handles[segment:segment+parts-1]
-        self._reversible_remove_handle(self._handles[segment])
-        if parts > 2:
-            self.merge_segment(segment, parts - 1)
-        else:
-            # Force orthogonal constraints to be recreated
-            self._update_orthogonal_constraints(self.orthogonal)
-        return deleted_handles
+        assert len(self._handles) >= 2, 'Not enough segments'
+        self._ports = []
+        handles = self._handles
+        for h1, h2 in zip(handles[:-1], handles[1:]):
+            self._ports.append(self._create_port(h1.pos, h2.pos))
 
 
     def opposite(self, handle):
@@ -673,11 +647,13 @@ class Line(Item):
         """
         super(Line, self).post_update(context)
         h0, h1 = self._handles[:2]
-        self._head_angle = atan2(h1.y - h0.y, h1.x - h0.x)
+        p0, p1 = h0.pos, h1.pos
+        self._head_angle = atan2(p1.y - p0.y, p1.x - p0.x)
         h1, h0 = self._handles[-2:]
-        self._tail_angle = atan2(h1.y - h0.y, h1.x - h0.x)
+        p1, p0 = h1.pos, h0.pos
+        self._tail_angle = atan2(p1.y - p0.y, p1.x - p0.x)
 
-    def closest_segment(self, x, y):
+    def closest_segment(self, pos):
         """
         Obtain a tuple (distance, point_on_line, segment).
         Distance is the distance from point to the closest line segment 
@@ -685,32 +661,30 @@ class Line(Item):
         Segment is the line segment closest to (x, y)
 
         >>> a = Line()
-        >>> a.closest_segment(4, 5)
+        >>> a.closest_segment((4, 5))
         (0.70710678118654757, (4.5, 4.5), 0)
         """
         h = self._handles
+        hpos = map(getattr, h, ['pos'] * len(h))
 
         # create a list of (distance, point_on_line) tuples:
-        distances = map(distance_line_point, h[:-1], h[1:], [(x, y)] * (len(h) - 1))
+        distances = map(distance_line_point, hpos[:-1], hpos[1:], [pos] * (len(hpos) - 1))
         distances, pols = zip(*distances)
         return reduce(min, zip(distances, pols, range(len(distances))))
 
-    def point(self, x, y):
+    def point(self, pos):
         """
         >>> a = Line()
-        >>> a.handles()[1].pos = 30, 30
-        >>> a.split_segment(0)
-        [<Handle object on (15, 15)>]
         >>> a.handles()[1].pos = 25, 5
-        >>> a.point(-1, 0)
+        >>> a._handles.append(a._create_handle((30, 30)))
+        >>> a.point((-1, 0))
         1.0
-        >>> '%.3f' % a.point(5, 4)
+        >>> '%.3f' % a.point((5, 4))
         '2.942'
-        >>> '%.3f' % a.point(29, 29)
+        >>> '%.3f' % a.point((29, 29))
         '0.784'
         """
-        h = self._handles
-        distance, point, segment = self.closest_segment(x, y)
+        distance, point, segment = self.closest_segment(pos)
         return max(0, distance - self.fuzziness)
 
     def draw_head(self, context):
@@ -731,11 +705,11 @@ class Line(Item):
         Draw the line itself.
         See Item.draw(context).
         """
-        def draw_line_end(handle, angle, draw):
+        def draw_line_end(pos, angle, draw):
             cr = context.cairo
             cr.save()
             try:
-                cr.translate(handle.x, handle.y)
+                cr.translate(*pos)
                 cr.rotate(angle)
                 draw(context)
             finally:
@@ -743,13 +717,20 @@ class Line(Item):
 
         cr = context.cairo
         cr.set_line_width(self.line_width)
-        draw_line_end(self._handles[0], self._head_angle, self.draw_head)
+        draw_line_end(self._handles[0].pos, self._head_angle, self.draw_head)
         for h in self._handles[1:-1]:
-            cr.line_to(h.x, h.y)
-        h0, h1 = self._handles[-2:]
-        draw_line_end(self._handles[-1], self._tail_angle, self.draw_tail)
+            cr.line_to(*h.pos)
+        draw_line_end(self._handles[-1].pos, self._tail_angle, self.draw_tail)
         cr.stroke()
 
+        ### debug code to draw line ports
+        ### cr.set_line_width(1)
+        ### cr.set_source_rgb(1.0, 0.0, 0.0)
+        ### for p in self.ports():
+        ###     cr.move_to(*p.start)
+        ###     cr.line_to(*p.end)
+        ### cr.stroke()
+
 
 
 __test__ = {
@@ -757,8 +738,4 @@ __test__ = {
     }
 
 
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()
-
 # vim: sw=4:et:ai
