@@ -33,15 +33,15 @@ method) changing appropriate variables to make the constraint valid again.
 
 from __future__ import division
 
-__version__ = "$Revision: 2093 $"
-# $HeadURL: http://svn.devjavu.com/gaphor/gaphas/tags/gaphas-0.3.6/gaphas/solver.py $
+__version__ = "$Revision$"
+# $HeadURL$
 
 from operator import isCallable
 from state import observed, reversible_pair, reversible_property
 
 # epsilon for float comparison
 # is simple abs(x - y) > EPSILON enough for canvas needs?
-EPSILON = 1e-10
+EPSILON = 1e-6
 
 # Variable Strengths:
 VERY_WEAK = 0
@@ -95,8 +95,11 @@ class Variable(object):
 
     @observed
     def set_value(self, value):
-        self._value = float(value)
-        self.dirty()
+        oldval = self._value
+        if abs(oldval - value) > EPSILON:
+            #print id(self), oldval, value
+            self._value = float(value)
+            self.dirty()
 
     value = reversible_property(lambda s: s._value, set_value)
 
@@ -345,7 +348,7 @@ class Projection(object):
 
     value = property(lambda s: s._var.value, _set_value)
 
-    strength = property(lambda s: s.variable().strength)
+    strength = property(lambda s: s._var.strength)
 
     def variable(self):
         """
@@ -373,6 +376,9 @@ class Solver(object):
         self._marked_cons = []
         self._solving = False
 
+    constraints = property(lambda s: s._constraints)
+
+
     def request_resolve(self, variable, projections_only=False):
         """
         Mark a variable as "dirty". This means it it solved the next time
@@ -399,11 +405,14 @@ class Solver(object):
         Variable(2, 20)
         >>> b.value=2.0
         >>> c_eq.weakest()
-        Variable(5, 20)
+        Variable(2, 20)
         >>> a.value=5.0
         >>> c_eq.weakest()
         Variable(2, 20)
         """
+        # Peel of Projections:
+        while isinstance(variable, Projection):
+            variable = variable.variable()
         for c in variable._constraints:
             if not projections_only or c._solver_has_projections:
                 if not self._solving:
@@ -416,7 +425,7 @@ class Solver(object):
                     self._marked_cons.append(c)
                     if __debug__:
                         if self._marked_cons.count(c) > 100:
-                            raise JuggleError, 'Variable juggling detected, constraint %s' % c
+                            raise JuggleError, 'Variable juggling detected, constraint %s resolved %d times out of %d' % (c, self._marked_cons.count(c), len(self._marked_cons))
 
 
     @observed
@@ -616,6 +625,7 @@ class Solver(object):
         finally:
             self._solving = False
 
+
 class solvable(object):
     """
     Easy-to-use drop Variable descriptor.
@@ -674,8 +684,4 @@ __test__ = {
     }
 
 
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()
-
 # vim:sw=4:et:ai
