@@ -2,8 +2,8 @@
 This module contains everything to display a Canvas on a screen.
 """
 
-__version__ = "$Revision: 2173 $"
-# $HeadURL: http://svn.devjavu.com/gaphor/gaphas/tags/gaphas-0.3.6/gaphas/view.py $
+__version__ = "$Revision$"
+# $HeadURL$
 
 import gobject
 import gtk
@@ -11,17 +11,16 @@ from cairo import Matrix
 from canvas import Context
 from geometry import Rectangle
 from quadtree import Quadtree
-from tool import ToolContext, DefaultTool
+from tool import DefaultTool
 from painter import DefaultPainter, BoundingBoxPainter
 from decorators import async, PRIORITY_HIGH_IDLE
 from decorators import nonrecursive
-from gaphas.sorter import Sorted
 
 # Handy debug flag for drawing bounding boxes around the items.
 DEBUG_DRAW_BOUNDING_BOX = False
 DEBUG_DRAW_QUADTREE = False
 
-# The default cursor (in case of a cursor reset)
+# The default cursor (use in case of a cursor reset)
 DEFAULT_CURSOR = gtk.gdk.LEFT_PTR
 
 
@@ -35,12 +34,15 @@ class View(object):
         self._painter = DefaultPainter()
 
         # Handling selections.
-        self._selected_items = Sorted(canvas)
+        ### TODO: Move this to a context?
+        self._selected_items = set()
         self._focused_item = None
         self._hovered_item = None
         self._dropzone_item = None
+        ###/
 
         self._qtree = Quadtree()
+        self._bounds = Rectangle(0, 0, 0, 0)
 
         self._canvas = None
         if canvas:
@@ -57,22 +59,31 @@ class View(object):
         in the view.
         """
         if self._canvas:
-            self._qtree = Quadtree()
+            self._qtree.clear()
+            self._selected_items.clear()
+            self._focused_item = None
+            self._hovered_item = None
+            self._dropzone_item = None
 
         self._canvas = canvas
-        self._selected_items.canvas = canvas
-        
 
     canvas = property(lambda s: s._canvas, _set_canvas)
 
 
-    def emit(self, args, **kwargs):
+    def emit(self, *args, **kwargs):
         """
         Placeholder method for signal emission functionality.
         """
         pass
 
 
+    def queue_draw_item(self, *items):
+        """
+        Placeholder for item redraw queueing.
+        """
+        pass
+
+
     def select_item(self, item):
         """
         Select an item. This adds @item to the set of selected items. Do::
@@ -126,7 +137,7 @@ class View(object):
             self.queue_draw_item(self._focused_item, item)
 
         if item:
-            self.selected_items = item #.add(item)
+            self.select_item(item)
         if item is not self._focused_item:
             self._focused_item = item
             self.emit('focus-changed', item)
@@ -136,7 +147,7 @@ class View(object):
         """
         Items that loose focus remain selected.
         """
-        self.focused_item = None
+        self._set_focused_item(None)
         
 
     focused_item = property(lambda s: s._focused_item,
@@ -148,9 +159,8 @@ class View(object):
         """
         Set the hovered item.
         """
-        if not item is self._hovered_item:
-            self.queue_draw_item(self._hovered_item, item)
         if item is not self._hovered_item:
+            self.queue_draw_item(self._hovered_item, item)
             self._hovered_item = item
             self.emit('hover-changed', item)
 
@@ -159,7 +169,7 @@ class View(object):
         """
         Unset the hovered item.
         """
-        self.hovered_item = None
+        self._set_hovered_item(None)
         
 
     hovered_item = property(lambda s: s._hovered_item,
@@ -181,7 +191,7 @@ class View(object):
         """
         Unset dropzone item.
         """
-        self._dropzone_item = None
+        self._set_dropzone_item(None)
 
 
     dropzone_item = property(lambda s: s._dropzone_item,
@@ -200,26 +210,122 @@ class View(object):
     painter = property(lambda s: s._painter, _set_painter)
 
 
-    def get_item_at_point(self, x, y, selected=True):
+    def get_item_at_point(self, pos, selected=True):
         """
-        Return the topmost item located at (x, y).
+        Return the topmost item located at ``pos`` (x, y).
 
         Parameters:
          - selected: if False returns first non-selected item
         """
-        point = (x, y)
-        items = self._qtree.find_intersect((x, y, 1, 1))
-        for item in self._canvas.sorter.sort(items, reverse=True):
+        items = self._qtree.find_intersect((pos[0], pos[1], 1, 1))
+        for item in self._canvas.sort(items, reverse=True):
             if not selected and item in self.selected_items:
                 continue  # skip selected items
 
             v2i = self.get_matrix_v2i(item)
-            ix, iy = v2i.transform_point(x, y)
-            if item.point(ix, iy) < 0.5:
+            ix, iy = v2i.transform_point(*pos)
+            if item.point((ix, iy)) < 0.5:
                 return item
         return None
 
 
+    def get_handle_at_point(self, pos, distance=6):
+        """
+        Look for a handle at ``pos`` and return the
+        tuple (item, handle).
+        """
+        def find(item):
+            """ Find item's handle at pos """
+            v2i = self.get_matrix_v2i(item)
+            d = v2i.transform_distance(distance, 0)[0]
+            x, y = v2i.transform_point(*pos)
+
+            for h in item.handles():
+                if not h.movable:
+                    continue
+                hx, hy = h.pos
+                if -d < (hx - x) < d and -d < (hy - y) < d:
+                    return h
+
+        # The focused item is the prefered item for handle grabbing
+        if self.focused_item:
+            h = find(self.focused_item)
+            if h:
+                return self.focused_item, h
+
+        # then try hovered item
+        if self.hovered_item:
+            h = find(self.hovered_item)
+            if h:
+                return self.hovered_item, h
+
+        # Last try all items, checking the bounding box first
+        x, y = pos
+        items = self.get_items_in_rectangle((x - distance, y - distance, distance * 2, distance * 2), reverse=True)
+
+        found_item, found_h = None, None
+        for item in items:
+            h = find(item)
+            if h:
+                return item, h
+        return None, None
+
+
+    def get_port_at_point(self, vpos, distance=10, exclude=None):
+        """
+        Find item with port closest to specified position.
+
+        List of items to be ignored can be specified with `exclude`
+        parameter.
+
+        Tuple is returned
+
+        - found item
+        - closest, connectable port
+        - closest point on found port (in view coordinates)
+
+        :Parameters:
+         vpos
+            Position specified in view coordinates.
+         distance
+            Max distance from point to a port (default 10)
+         exclude
+            Set of items to ignore.
+        """
+        v2i = self.get_matrix_v2i
+        vx, vy = vpos
+
+        max_dist = distance
+        port = None
+        glue_pos = None
+        item = None
+
+        rect = (vx - distance, vy - distance, distance * 2, distance * 2)
+        items = self.get_items_in_rectangle(rect, reverse=True)
+        for i in items:
+            if i in exclude:
+                continue
+            for p in i.ports():
+                if not p.connectable:
+                    continue
+
+                ix, iy = v2i(i).transform_point(vx, vy)
+                pg, d = p.glue((ix, iy))
+
+                if d >= max_dist:
+                    continue
+
+                item = i
+                port = p
+
+                # transform coordinates from connectable item space to view
+                # space
+                i2v = self.get_matrix_i2v(i).transform_point
+                glue_pos = i2v(*pg)
+
+        return item, port, glue_pos
+
+
     def get_items_in_rectangle(self, rect, intersect=True, reverse=False):
         """
         Return the items in the rectangle 'rect'.
@@ -229,7 +335,7 @@ class View(object):
             items = self._qtree.find_intersect(rect)
         else:
             items = self._qtree.find_inside(rect)
-        return self._canvas.sorter.sort(items, reverse=reverse)
+        return self._canvas.sort(items, reverse=reverse)
 
 
     def select_in_rectangle(self, rect):
@@ -245,16 +351,19 @@ class View(object):
         """
         Zoom in/out by factor @factor.
         """
+        # TODO: should the scale factor be clipped?
         self._matrix.scale(factor, factor)
 
         # Make sure everything's updated
-        map(self.update_matrix, self._canvas.get_all_items())
-        self.request_update(self._canvas.get_all_items())
+        #map(self.update_matrix, self._canvas.get_all_items())
+        self.request_update((), self._canvas.get_all_items())
 
 
     def set_item_bounding_box(self, item, bounds):
         """
-        Update the bounding box of the item (in canvas coordinates).
+        Update the bounding box of the item.
+
+        ``bounds`` is in view coordinates.
 
         Coordinates are calculated back to item coordinates, so matrix-only
         updates can occur.
@@ -267,12 +376,12 @@ class View(object):
 
     def get_item_bounding_box(self, item):
         """
-        Get the bounding box for the item, in canvas coordinates.
+        Get the bounding box for the item, in view coordinates.
         """
         return self._qtree.get_bounds(item)
 
 
-    bounding_box = property(lambda s: Rectangle(*s._qtree.soft_bounds))
+    bounding_box = property(lambda s: s._bounds)
 
 
     def update_bounding_box(self, cr, items=None):
@@ -322,13 +431,19 @@ class View(object):
         """
         Update item matrices related to view.
         """
-        i2v = item._matrix_i2c * self._matrix
+        try:
+            i2v = item._matrix_i2c.multiply(self._matrix)
+        except AttributeError:
+            # Fall back to old behaviour
+            i2v = item._matrix_i2c * self._matrix
+
         item._matrix_i2v[self] = i2v
 
         v2i = Matrix(*i2v)
         v2i.invert()
         item._matrix_v2i[self] = v2i
 
+
     def _clear_matrices(self):
         """
         Clear registered data in Item's _matrix{i2c|v2i} attributes.
@@ -341,17 +456,6 @@ class View(object):
                 pass
 
 
-# Map GDK events to tool methods
-EVENT_HANDLERS = {
-    gtk.gdk.BUTTON_PRESS: 'on_button_press',
-    gtk.gdk.BUTTON_RELEASE: 'on_button_release',
-    gtk.gdk._2BUTTON_PRESS: 'on_double_click',
-    gtk.gdk._3BUTTON_PRESS: 'on_triple_click',
-    gtk.gdk.MOTION_NOTIFY: 'on_motion_notify',
-    gtk.gdk.KEY_PRESS: 'on_key_press',
-    gtk.gdk.KEY_RELEASE: 'on_key_release'
-}
-
 
 
 class GtkView(gtk.DrawingArea, View):
@@ -372,6 +476,8 @@ class GtkView(gtk.DrawingArea, View):
     
     # Signals: emited after the change takes effect.
     __gsignals__ = {
+        'set-scroll-adjustments': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
+                      (gtk.Adjustment, gtk.Adjustment)),
         'dropzone-changed': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                       (gobject.TYPE_PYOBJECT,)),
         'hover-changed': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
@@ -387,7 +493,7 @@ class GtkView(gtk.DrawingArea, View):
     }
 
 
-    def __init__(self, canvas=None):
+    def __init__(self, canvas=None, hadjustment=None, vadjustment=None):
         gtk.DrawingArea.__init__(self)
 
         self._dirty_items = set()
@@ -400,21 +506,18 @@ class GtkView(gtk.DrawingArea, View):
                         | gtk.gdk.BUTTON_RELEASE_MASK
                         | gtk.gdk.POINTER_MOTION_MASK
                         | gtk.gdk.KEY_PRESS_MASK
-                        | gtk.gdk.KEY_RELEASE_MASK)
+                        | gtk.gdk.KEY_RELEASE_MASK
+                        | gtk.gdk.SCROLL_MASK)
 
-        self._hadjustment = gtk.Adjustment()
-        self._vadjustment = gtk.Adjustment()
-        self._hadjustment.connect('value-changed', self.on_adjustment_changed)
-        self._vadjustment.connect('value-changed', self.on_adjustment_changed)
+        self._hadjustment = None
+        self._vadjustment = None
+        self._hadjustment_handler_id = None
+        self._vadjustment_handler_id = None
 
-        self._tool = DefaultTool()
+        self.emit('set-scroll-adjustments', hadjustment, vadjustment)
+
+        self._set_tool(DefaultTool())
         
-        # TODO: add some "tool" to do some low priority event post-processing
-        #    This way we can add features like element alignment in a fairly
-        #    non-intrusive way. Add a few methods that allow tools to mark
-        #    item/handle pairs for post-processing.
-        #    maybe adding it to the tool context would be enough (?).
-
         # Set background to white.
         self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#FFF'))
 
@@ -442,7 +545,7 @@ class GtkView(gtk.DrawingArea, View):
         if self._canvas:
             self._canvas.register_view(self)
             self.request_update(self._canvas.get_all_items())
-
+        self.queue_draw_refresh()
 
     canvas = property(lambda s: s._canvas, _set_canvas)
 
@@ -452,6 +555,7 @@ class GtkView(gtk.DrawingArea, View):
         Set the tool to use. Tools should implement tool.Tool.
         """
         self._tool = tool
+        tool.set_view(self)
         self.emit('tool-changed')
 
 
@@ -463,6 +567,26 @@ class GtkView(gtk.DrawingArea, View):
 
     vadjustment = property(lambda s: s._vadjustment)
 
+    
+    def do_set_scroll_adjustments(self, hadjustment, vadjustment):
+        if self._hadjustment_handler_id:
+            self._hadjustment.disconnect(self._hadjustment_handler_id)
+            self._hadjustment_handler_id = None
+        if self._vadjustment_handler_id:
+            self._vadjustment.disconnect(self._vadjustment_handler_id)
+            self._vadjustment_handler_id = None
+
+        self._hadjustment = hadjustment or gtk.Adjustment()
+        self._vadjustment = vadjustment or gtk.Adjustment()
+
+        self._hadjustment_handler_id = \
+                        self._hadjustment.connect('value-changed',
+                                                  self.on_adjustment_changed)
+        self._vadjustment_handler_id = \
+                        self._vadjustment.connect('value-changed',
+                                                  self.on_adjustment_changed)
+        self.update_adjustments()
+
 
     def zoom(self, factor):
         """
@@ -472,74 +596,80 @@ class GtkView(gtk.DrawingArea, View):
         self.queue_draw_refresh()
 
 
-    def _update_adjustment(self, adjustment, value, canvas_size, canvas_offset, viewport_size):
-        """
-        >>> v = GtkView()
-        >>> a = gtk.Adjustment()
-        >>> v._hadjustment = a
-        >>> v._update_adjustment(a, 10, 100, 0, 20)
-        >>> a.page_size, a.page_increment, a.value
-        (20.0, 20.0, 0.0)
-        """
-        #canvas_size += viewport_size
-        #canvas_offset -= viewport_size
-        if viewport_size != adjustment.page_size or canvas_size != adjustment.upper:
-            adjustment.page_size = viewport_size
-            adjustment.page_increment = viewport_size
-            adjustment.step_increment = viewport_size/10
-            adjustment.upper = adjustment.value + canvas_offset + canvas_size
-            adjustment.lower = adjustment.value + canvas_offset
-        
-        if adjustment.value > adjustment.upper - viewport_size:
-            adjustment.value = adjustment.upper - viewport_size
-
     @async(single=True)
     def update_adjustments(self, allocation=None):
-        """
-        Update the allocation objects (for scrollbars).
-        """
         if not allocation:
             allocation = self.allocation
 
-        # Define a minimal view size:
-        aw, ah = allocation.width, allocation.height 
-        x, y, w, h = self.bounding_box + (self.matrix.transform_point(0, 0) + (aw * 2, ah * 2))
-        self._update_adjustment(self._hadjustment,
-                                value = self._hadjustment.value,
-                                canvas_size=w,
-                                canvas_offset=x,
-                                viewport_size=allocation.width)
-        self._update_adjustment(self._vadjustment,
-                                value = self._vadjustment.value,
-                                canvas_size=h,
-                                canvas_offset=y,
-                                viewport_size=allocation.height)
+        hadjustment = self._hadjustment
+        vadjustment = self._vadjustment
+
+        # canvas limits (in view coordinates)
+        c = Rectangle(*self._qtree.soft_bounds)
+
+        # view limits
+        v = Rectangle(0, 0, self.allocation.width, self.allocation.height)
+
+        # union of these limits gives scrollbar limits
+        if v in c:
+            u = c
+        else:
+            u = c + v
+
+        # set lower limits
+        hadjustment.lower, vadjustment.lower = u.x, u.y
+
+        # set upper limits
+        hadjustment.upper, vadjustment.upper = u.x1, u.y1
+
+        # set page size
+        aw, ah = self.allocation.width, self.allocation.height
+        hadjustment.page_size = aw
+        vadjustment.page_size = ah
+
+        # set increments
+        hadjustment.page_increment = aw
+        hadjustment.step_increment = aw / 10
+        vadjustment.page_increment = ah
+        vadjustment.step_increment = ah / 10
+
+        # set position
+        if v.x != hadjustment.value or v.y != vadjustment.value:
+            hadjustment.value, vadjustment.value = v.x, v.y
 
-        x, y, w, h = self._qtree.bounds
-        if w != aw or h != ah:
-            self._qtree.resize((0, 0, aw, ah))
-        
 
     def queue_draw_item(self, *items):
         """
         Like ``DrawingArea.queue_draw_area``, but use the bounds of the
         item as update areas. Of course with a pythonic flavor: update
         any number of items at once.
+
+        TODO: Should we also create a (sorted) list of items that need redrawal?
         """
-        queue_draw_area = self.queue_draw_area
         get_bounds = self._qtree.get_bounds
-        for item in items:
-            try:
-                queue_draw_area(*get_bounds(item))
-            except KeyError:
-                pass # No bounds calculated yet? bummer.
+        items = filter(None, items)
+        try:
+            # create a copy, otherwise we'll change the original rectangle
+            bounds = Rectangle(*get_bounds(items[0]))
+            for item in items[1:]:
+                bounds += get_bounds(item)
+            self.queue_draw_area(*bounds)
+        except IndexError:
+            pass
+        except KeyError:
+            pass # No bounds calculated yet? bummer.
 
 
     def queue_draw_area(self, x, y, w, h):
         """
         Wrap draw_area to convert all values to ints.
         """
-        super(GtkView, self).queue_draw_area(int(x), int(y), int(w+1), int(h+1))
+        try:
+            super(GtkView, self).queue_draw_area(int(x), int(y), int(w+1), int(h+1))
+        except OverflowError:
+            # Okay, now the zoom factor is very large or something
+            a = self.allocation
+            super(GtkView, self).queue_draw_area(0, 0, a.width, a.height)
 
 
     def queue_draw_refresh(self):
@@ -589,18 +719,16 @@ class GtkView(gtk.DrawingArea, View):
         dirty_matrix_items = self._dirty_matrix_items
 
         try:
-            for i in dirty_items:
-                self.queue_draw_item(i)
+            self.queue_draw_item(*dirty_items)
 
+            # Mark old bb section for update
+            self.queue_draw_item(*dirty_matrix_items)
             for i in dirty_matrix_items:
                 if i not in self._qtree:
                     dirty_items.add(i)
                     self.update_matrix(i)
                     continue
 
-                # Mark old bb section for update
-                self.queue_draw_item(i)
-
                 self.update_matrix(i)
 
                 if i not in dirty_items:
@@ -613,7 +741,7 @@ class GtkView(gtk.DrawingArea, View):
                     vbounds = Rectangle(x0, y0, x1=x1, y1=y1)
                     self._qtree.add(i, vbounds, bounds)
 
-                self.queue_draw_item(i)
+            self.queue_draw_item(*dirty_matrix_items)
 
             # Request bb recalculation for all 'really' dirty items
             self.update_bounding_box(set(dirty_items))
@@ -652,7 +780,6 @@ class GtkView(gtk.DrawingArea, View):
 
     def do_realize(self):
         gtk.DrawingArea.do_realize(self)
-        allocation = self.allocation
 
         if self._canvas:
             self.request_update(self._canvas.get_all_items())
@@ -663,10 +790,10 @@ class GtkView(gtk.DrawingArea, View):
             # (weak refs), better do it explicitly to be sure.
             self._clear_matrices()
             self.canvas = None
-        self._qtree = None
+        self._qtree.clear()
 
-        self._dirty_items = None
-        self._dirty_matrix_items = None
+        self._dirty_items.clear()
+        self._dirty_matrix_items.clear()
 
         gtk.DrawingArea.do_unrealize(self)
 
@@ -719,9 +846,8 @@ class GtkView(gtk.DrawingArea, View):
         """
         Handle GDK events. Events are delegated to a `tool.Tool`.
         """
-        handler = EVENT_HANDLERS.get(event.type)
-        if self._tool and handler:
-            return getattr(self._tool, handler)(ToolContext(view=self), event) and True or False
+        if self._tool:
+            return self._tool.handle(event) and True or False
         return False
 
 
@@ -730,20 +856,32 @@ class GtkView(gtk.DrawingArea, View):
         Change the transformation matrix of the view to reflect the
         value of the x/y adjustment (scrollbar).
         """
+        if adj.value == 0.0: return
+
+        # Can not use self._matrix.translate( - adj.value , 0) here, since
+        # the translate method effectively does a m * self._matrix, which
+        # will result in the translation being multiplied by the orig. matrix
+
+        m = Matrix()
         if adj is self._hadjustment:
-            self._matrix.translate(- self._matrix[4] / self._matrix[0] - adj.value , 0)
+            m.translate( - adj.value, 0)
         elif adj is self._vadjustment:
-            self._matrix.translate(0, - self._matrix[5] / self._matrix[3] - adj.value)
+            m.translate(0, - adj.value)
+        self._matrix *= m
 
         # Force recalculation of the bounding boxes:
-        map(self.update_matrix, self._canvas.get_all_items())
         self.request_update((), self._canvas.get_all_items())
 
         self.queue_draw_refresh()
 
 
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()
+# Set a signal to set adjustments. This way a ScrolledWindow can set its own
+# Adjustment objects on the View. Otherwise a warning is shown:
+#
+#   GtkWarning: gtk_scrolled_window_add(): cannot add non scrollable widget
+#   use gtk_scrolled_window_add_with_viewport() instead
 
-# vim: sw=4:et:
+GtkView.set_set_scroll_adjustments_signal("set-scroll-adjustments")
+
+
+# vim: sw=4:et:ai
